import sys

import click

from hapless import Hap, Hapless, config

hapless = Hapless(hapless_dir=config.HAPLESS_DIR)
console = hapless.ui


def get_or_exit(hap_alias: str) -> Hap:
    hap = hapless.get_hap(hap_alias)
    if hap is None:
        console.error(f"No such hap: {hap_alias}")
        sys.exit(1)
    if not hap.accessible:
        console.error(f"Cannot manage hap launched by another user. Owner: {hap.owner}")
        sys.exit(1)
    return hap


hap_argument = click.argument(
    "hap_alias",
    metavar="hap",
    required=True,
)
hap_argument_optional = click.argument(
    "hap_alias",
    metavar="hap",
    required=False,
)



try:
    from importlib.metadata import version
except ModuleNotFoundError:
    # Fallback for Python 3.7
    from importlib_metadata import version

import abc
import json
from itertools import filterfalse
from typing import List

from rich import box
from rich.console import Group, RenderableType
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from hapless import config
from hapless.hap import Hap, Status


class Formatter(abc.ABC):
    """
    Abstract base class for formatting Hap objects.
    """

    def __init__(self, verbose: bool = False):
        self.verbose = verbose

    @abc.abstractmethod
    def format_one(self, hap: Hap) -> RenderableType:
        pass

    @abc.abstractmethod
    def format_list(self, haps: List[Hap]) -> RenderableType:
        pass


class TableFormatter(Formatter):
    """
    Formats Hap objects as a rich table.
    """

    def _get_status_text(self, status: Status) -> Text:
        color = config.STATUS_COLORS.get(status, "dim")
        status_text = Text()
        status_text.append(config.ICON_STATUS, style=color)
        status_text.append(f" {status.value}")
        return status_text

    def format_one(self, hap: Hap) -> Group:
        status_table = Table(show_header=False, show_footer=False, box=box.SIMPLE)

        status_text = self._get_status_text(hap.status)
        status_table.add_row("Status:", status_text)

        status_table.add_row("PID:", f"{hap.pid or '-'}")

        if hap.rc is not None:
            status_table.add_row("Return code:", f"{hap.rc}")

        cmd_text = Text(f"{hap.cmd}", style=f"{config.COLOR_ACCENT} bold")
        if self.verbose:
            status_table.add_row("")
            status_table.add_row("Command:", cmd_text)
            status_table.add_row("Working dir:", f"{hap.workdir}")
            status_table.add_row("")
        else:
            status_table.add_row("Command:", cmd_text)

        proc = hap.proc
        if self.verbose and proc is not None:
            status_table.add_row("Parent PID:", f"{proc.ppid()}")
            status_table.add_row("User:", f"{proc.username()}")

        if self.verbose and hap.redirect_stderr:
            # Paths are the same, show one line only
            status_table.add_row("Logs file:", f"{hap.stdout_path}")

        if self.verbose and not hap.redirect_stderr:
            status_table.add_row("Stdout file:", f"{hap.stdout_path}")
            status_table.add_row("Stderr file:", f"{hap.stderr_path}")

        start_time = hap.start_time
        end_time = hap.end_time
        if self.verbose and start_time:
            status_table.add_row("Start time:", f"{start_time}")

        if self.verbose and end_time:
            status_table.add_row("End time:", f"{end_time}")

        status_table.add_row("Runtime:", f"{hap.runtime}")

        status_panel = Panel(
            status_table,
            expand=self.verbose,
            title=f"Hap {config.ICON_HAP}{hap.hid}",
            subtitle=hap.name,
        )
        result = Group(status_panel)

        environ = hap.env
        if self.verbose and environ is not None:
            env_table = Table(show_header=False, show_footer=False, box=None)
            env_table.add_column("", justify="right")
            env_table.add_column("", justify="left", style=config.COLOR_ACCENT)

            for key, value in environ.items():
                env_table.add_row(key, Text(value, overflow="fold"))

            env_panel = Panel(
                env_table,
                title="Environment",
                subtitle=f"{len(environ)} items",
                border_style=config.COLOR_MAIN,
            )
            result = Group(status_panel, env_panel)
        return result

    def format_list(self, haps: List[Hap]) -> Table:
        package_version = version(__package__)
        table = Table(
            show_header=True,
            header_style=f"{config.COLOR_MAIN} bold",
            box=box.HEAVY_EDGE,
            caption_style="dim",
            caption_justify="right",
        )
        table.add_column("#", style="dim", min_width=2)
        table.add_column("Name")
        table.add_column("PID")
        if self.verbose:
            table.add_column("Command")
            table.add_column("Owner")
        table.add_column("Status")
        table.add_column("RC", justify="right")
        table.add_column("Runtime", justify="right")

        active_haps = 0
        for hap in haps:
            active_haps += 1 if hap.active else 0
            name = Text(hap.name)
            if hap.restarts:
                name += Text(f"{config.RESTART_DELIM}{hap.restarts}", style="dim")
            pid_text = (
                f"{hap.pid}" if hap.active else Text(f"{hap.pid or '-'}", style="dim")
            )
            command_text = Text(
                hap.cmd, overflow="ellipsis", style=f"{config.COLOR_ACCENT}"
            )
            status_text = self._get_status_text(hap.status)
            command_text.truncate(config.TRUNCATE_LENGTH)
            row = [
                f"{hap.hid}",
                name,
                pid_text,
                command_text if self.verbose else None,
                hap.owner if self.verbose else None,
                status_text,
                f"{hap.rc}" if hap.rc is not None else "",
                hap.runtime,
            ]
            table.add_row(*filterfalse(lambda x: x is None, row))

        if self.verbose:
            table.title = f"{config.ICON_HAP} {__package__}, {package_version}"
            table.caption = f"{active_haps} active / {len(haps)} total"

        return table


class JSONFormatter(Formatter):
    """
    Formats Hap objects as a valid JSON.
    """

    def format_one(self, hap: Hap) -> str:
        return json.dumps(hap.serialize())

    def format_list(self, haps: List[Hap]) -> str:
        return json.dumps([hap.serialize() for hap in haps])



import sys
from typing import Optional

import click

from hapless.cli_utils import (
    console,
    get_or_exit,
    hap_argument,
    hap_argument_optional,
    hapless,
)
from hapless.formatters import JSONFormatter, TableFormatter
from hapless.utils import validate_signal

try:
    from shlex import join as shlex_join
except ImportError:
    # Fallback for Python 3.7
    from hapless.utils import shlex_join_backport as shlex_join


@click.group(invoke_without_command=True)
@click.version_option(message="hapless, version %(version)s")
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.option(
    "--json", "json_output", is_flag=True, default=False, help="Output in JSON format."
)
@click.pass_context
def cli(ctx, verbose: bool, json_output: bool):
    if ctx.invoked_subcommand is None:
        _status(None, verbose=verbose, json_output=json_output)


@cli.command(short_help="Display information about haps.")
@hap_argument_optional
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.option(
    "--json", "json_output", is_flag=True, default=False, help="Output in JSON format."
)
def status(hap_alias: Optional[str], verbose: bool, json_output: bool):
    _status(hap_alias, verbose, json_output=json_output)


@cli.command(short_help="Same as a status.")
@hap_argument_optional
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.option(
    "--json", "json_output", is_flag=True, default=False, help="Output in JSON format."
)
def show(hap_alias: Optional[str], verbose: bool, json_output: bool):
    _status(hap_alias, verbose, json_output=json_output)


def _status(
    hap_alias: Optional[str] = None,
    verbose: bool = False,
    json_output: bool = False,
):
    formatter_cls = JSONFormatter if json_output else TableFormatter
    formatter = formatter_cls(verbose=verbose)
    if hap_alias is not None:
        hap = get_or_exit(hap_alias)
        hapless.show(hap, formatter=formatter)
    else:
        haps = hapless.get_haps(accessible_only=False)
        hapless.stats(haps, formatter=formatter)


@cli.command(short_help="Output logs for a hap.")
@hap_argument
@click.option("-f", "--follow", is_flag=True, default=False)
@click.option("-e", "--stderr", is_flag=True, default=False)
def logs(hap_alias: str, follow, stderr):
    hap = get_or_exit(hap_alias)
    hapless.logs(hap, stderr=stderr, follow=follow)


@cli.command(short_help="Remove successfully completed haps.")
@click.option(
    "-a",
    "--all",
    "clean_all",
    is_flag=True,
    default=False,
    help="Include failed haps for the removal.",
)
def clean(clean_all):
    hapless.clean(clean_all=clean_all)


@cli.command(short_help="Remove all finished haps, including failed ones.")
def cleanall():
    hapless.clean(clean_all=True)


@cli.command(
    short_help="Execute background command as a hap.",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.argument("cmd", nargs=-1)
@click.option(
    "-n", "--name", help="Provide your own alias for the hap instead of a default one."
)
@click.option(
    "--check",
    is_flag=True,
    default=False,
    help="Verify command launched does not fail immediately.",
)
def run(cmd, name, check):
    hap = hapless.get_hap(name)
    if hap is not None:
        console.error(f"Hap with such name already exists: {hap}")
        sys.exit(1)

    # NOTE: click doesn't like `required` property for `cmd` argument
    # https://click.palletsprojects.com/en/latest/arguments/#variadic-arguments
    cmd_escaped = shlex_join(cmd).strip()
    if not cmd_escaped:
        console.error("You have to provide a command to run")
        sys.exit(1)
    hapless.run_command(cmd_escaped, name=name, check=check)


@cli.command(short_help="Pause a specific hap.")
@hap_argument
def pause(hap_alias: str):
    hap = get_or_exit(hap_alias)
    hapless.pause_hap(hap)


@cli.command(short_help="Resume execution of a paused hap.")
@hap_argument
def resume(hap_alias: str):
    hap = get_or_exit(hap_alias)
    hapless.resume_hap(hap)


@cli.command(short_help="Terminate a specific hap / all haps.")
@hap_argument_optional
@click.option("-a", "--all", "killall", is_flag=True, default=False)
def kill(hap_alias: Optional[str], killall):
    if hap_alias is not None and killall:
        raise click.BadOptionUsage(
            "killall", "Cannot use --all flag while hap id provided"
        )

    if hap_alias is None and not killall:
        raise click.BadArgumentUsage("Provide hap alias to kill")

    if killall:
        haps = hapless.get_haps()
        hapless.kill(haps)
    else:
        hap = get_or_exit(hap_alias)
        hapless.kill([hap])


@cli.command(short_help="Send an arbitrary signal to a hap.")
@hap_argument
@click.argument("signal", callback=validate_signal, metavar="signal-code")
def signal(hap_alias: str, signal):
    hap = get_or_exit(hap_alias)
    hapless.signal(hap, signal)


@cli.command(short_help="Kills the hap and starts it again.")
@hap_argument
def restart(hap_alias: str):
    hap = get_or_exit(hap_alias)
    hapless.restart(hap)


@cli.command(short_help="Sets new name/alias for the existing hap.")
@hap_argument
@click.argument("new_name", metavar="new-name", required=True)
def rename(hap_alias: str, new_name: str):
    hap = get_or_exit(hap_alias)
    same_name_hap = hapless.get_hap(new_name)
    if same_name_hap is not None:
        console.print(f"Hap with such name already exists: {same_name_hap}")
        sys.exit(1)
    hapless.rename_hap(hap, new_name)


if __name__ == "__main__":
    cli()



import getpass
import os
import shutil
import signal
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Union

import psutil

from hapless import config
from hapless.formatters import Formatter
from hapless.hap import Hap, Status
from hapless.ui import ConsoleUI
from hapless.utils import kill_proc_tree, logger, wait_created


class Hapless:
    def __init__(
        self,
        hapless_dir: Optional[Union[Path, str]] = None,
        *,
        quiet: bool = False,
    ):
        self.ui = ConsoleUI(disable=quiet)
        user = getpass.getuser()
        default_dir = Path(tempfile.gettempdir()) / "hapless"

        hapless_dir = Path(hapless_dir or default_dir)
        try:
            if not hapless_dir.exists():
                hapless_dir.mkdir(parents=True, exist_ok=True)
            os.utime(hapless_dir)
        except PermissionError as e:
            logger.error(f"Cannot initialize state directory {hapless_dir}: {e}")
            self.ui.error(
                f"State directory {hapless_dir} is not accessible by user {user}"
            )
            sys.exit(1)

        self._hapless_dir = hapless_dir
        logger.debug(f"Initialized within {self._hapless_dir} dir")

    def stats(self, haps: List[Hap], formatter: Formatter):
        self.ui.stats(haps, formatter=formatter)

    def show(self, hap: Hap, formatter: Formatter):
        self.ui.show_one(hap, formatter=formatter)

    @property
    def dir(self) -> Path:
        return self._hapless_dir

    def _get_hap_dirs(self) -> List[str]:
        hap_dirs = filter(str.isdigit, os.listdir(self._hapless_dir))
        return sorted(hap_dirs, key=int)

    def _get_hap_names_map(self) -> Dict[str, str]:
        names = {}
        for dir in self._get_hap_dirs():
            filename = self._hapless_dir / dir / "name"
            if filename.exists():
                with open(filename) as f:
                    raw_name = f.read().strip()
                    name = raw_name.split(config.RESTART_DELIM)[0]
                    names[name] = dir
        return names

    def _get_next_hap_id(self) -> str:
        dirs = self._get_hap_dirs()
        next_num = 1 if not dirs else int(dirs[-1]) + 1
        return f"{next_num}"

    def get_hap(self, hap_alias: str) -> Optional[Hap]:
        dirs = self._get_hap_dirs()
        # Check by hap id
        if hap_alias in dirs:
            return Hap(self._hapless_dir / hap_alias)

        # Check by hap name
        names_map = self._get_hap_names_map()
        if hap_alias in names_map:
            return Hap(self._hapless_dir / names_map[hap_alias])

    def _get_all_haps(self) -> List[Hap]:
        """
        Get all haps available in the hapless state directory.
        Current user might only be able to access subset of them.
        """
        haps = []
        if not self._hapless_dir.exists():
            return haps

        for dir in self._get_hap_dirs():
            hap_path = self._hapless_dir / dir
            haps.append(Hap(hap_path))
        return haps

    def get_haps(self, accessible_only=True) -> List[Hap]:
        """
        Get all haps that are managable by the current user.
        If `accessible_only` is set to False, all haps will be returned.
        """

        def filter_haps(hap_arg: Hap) -> bool:
            return hap_arg.accessible if accessible_only else True

        haps = list(filter(filter_haps, self._get_all_haps()))
        return haps

    def create_hap(
        self,
        cmd: str,
        workdir: Optional[Union[str, Path]] = None,
        hid: Optional[str] = None,
        name: Optional[str] = None,
        *,
        redirect_stderr: Optional[bool] = None,
    ) -> Hap:
        hid = hid or self._get_next_hap_id()
        hap_dir = self._hapless_dir / f"{hid}"
        hap_dir.mkdir()
        if redirect_stderr is None:
            redirect_stderr = config.REDIRECT_STDERR
        if workdir is None or not Path(workdir).exists():
            workdir = os.getcwd()
        return Hap(
            hap_dir,
            name=name,
            cmd=cmd,
            workdir=workdir,
            redirect_stderr=redirect_stderr,
        )

    def _wrap_subprocess(self, hap: Hap):
        try:
            stdout_pipe = open(hap.stdout_path, "w")
            stderr_pipe = stdout_pipe
            if not hap.redirect_stderr:
                stderr_pipe = open(hap.stderr_path, "w")
            shell_exec = os.getenv("SHELL")
            if shell_exec is not None:
                logger.debug(f"Using {shell_exec} to run hap")
            proc = subprocess.Popen(
                hap.cmd,
                cwd=hap.workdir,
                shell=True,
                executable=shell_exec,
                stdout=stdout_pipe,
                stderr=stderr_pipe,
            )

            pid = proc.pid
            logger.debug(f"Attaching hap {hap} to pid {pid}")
            hap.bind(pid)

            retcode = proc.wait()
        finally:
            stdout_pipe.close()
            stderr_pipe.close()

        hap.set_return_code(retcode)

    def _check_fast_failure(self, hap: Hap) -> None:
        timeout = config.FAILFAST_TIMEOUT
        wait_created(
            hap._rc_file,
            live_context=self.ui.get_live(),
            interval=0.5,
            timeout=timeout,
        )
        return_code: Optional[int] = hap.rc
        if return_code is None:
            # no return code yet, process is still running
            self.ui.print(
                f"{config.ICON_INFO} Hap is healthy "
                f"and still running after {timeout} seconds",
                style=f"{config.COLOR_ACCENT} bold",
            )
        elif return_code == 0:
            # finished quickly, but successfully
            self.ui.print(
                f"{config.ICON_INFO} Hap finished successfully "
                f"in less than {timeout} seconds",
                style=f"{config.COLOR_ACCENT} bold",
            )
        else:
            # non-zero return code
            self.ui.error("Hap exited too quickly. stderr message:")
            self.ui.print(hap.stderr_path.read_text())
            sys.exit(1)

    def run_hap(
        self,
        hap: Hap,
        check: bool = False,
        *,
        blocking: bool = False,
    ) -> None:
        """
        Run hap in a separate process.
        If `check` is True, it will check for fast failure and exit
        if hap terminates too quickly.
        """
        if blocking:
            # NOTE: this is for the testing purposes only
            self._wrap_subprocess(hap)
            return

        self.ui.print(f"{config.ICON_INFO} Launching", hap)
        # TODO: or sys.platform == "win32"
        if config.NO_FORK:
            logger.debug("Forking is disabled, running using spawn via wrapper")
            self._run_via_spawn(hap)
        else:
            logger.debug("Running hap using fork")
            self._run_via_fork(hap)

        logger.debug(f"Parent process continues with pid {os.getpid()}")
        if check:
            self._check_fast_failure(hap)

    def _run_via_spawn(self, hap: Hap) -> None:
        exec_path = shutil.which("hapwrap")
        if exec_path is None:
            self.ui.error(
                "Cannot find wrapper to run process. Please reinstall hapless"
            )
            sys.exit(1)

        proc = subprocess.Popen(
            [f"{exec_path}", f"{hap.hid}"],
            start_new_session=True,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        logger.debug(f"Running subprocess in child with pid {proc.pid}")
        logger.debug(f"Using wrapper located at {exec_path}")

    def _run_via_fork(self, hap: Hap) -> None:
        pid = os.fork()
        if pid == 0:
            os.setsid()
            logger.debug(f"Running subprocess in child with pid {os.getpid()}")
            self._wrap_subprocess(hap)
            # NOTE: sole purpose of the child is to run a subprocess
            sys.exit(0)

    def run_command(
        self,
        cmd: str,
        workdir: Optional[Union[str, Path]] = None,
        hid: Optional[str] = None,
        name: Optional[str] = None,
        check: bool = False,
        *,
        redirect_stderr: Optional[bool] = None,
        blocking: bool = False,
    ) -> None:
        """
        For the command provided create a hap and run it.
        If `hid` or `name` is not provided, it will be generated automatically.
        """
        hap = self.create_hap(
            cmd=cmd,
            workdir=workdir,
            hid=hid,
            name=name,
            redirect_stderr=redirect_stderr,
        )
        self.run_hap(hap, check=check, blocking=blocking)

    def pause_hap(self, hap: Hap):
        proc = hap.proc
        if proc is not None:
            proc.suspend()
            self.ui.print(f"{config.ICON_INFO} Paused", hap)
        else:
            self.ui.error(f"Cannot pause. Hap {hap} is not running")
            sys.exit(1)

    def resume_hap(self, hap: Hap):
        proc = hap.proc
        if proc is not None and proc.status() == psutil.STATUS_STOPPED:
            proc.resume()
            self.ui.print(f"{config.ICON_INFO} Resumed", hap)
        else:
            self.ui.error(f"Cannot resume. Hap {hap} is not suspended")
            sys.exit(1)

    def logs(self, hap: Hap, stderr: bool = False, follow: bool = False):
        filepath = hap.stderr_path if stderr else hap.stdout_path
        if follow:
            self.ui.print(
                f"{config.ICON_INFO} Streaming {filepath} file...",
                style=f"{config.COLOR_MAIN} bold",
            )
            return subprocess.run(["tail", "-f", filepath])

        text = filepath.read_text()
        if not text:
            self.ui.error("No logs found")
            return

        self.ui.print(
            f"{config.ICON_INFO} Showing logs at {filepath}",
            style=f"{config.COLOR_MAIN} bold",
        )
        self.ui.print_plain(text)

    def _clean_haps(self, filter_haps) -> int:
        haps = list(filter(filter_haps, self.get_haps()))
        for hap in haps:
            logger.debug(f"Removing {hap.path}")
            shutil.rmtree(hap.path, ignore_errors=True)
        return len(haps)

    def _clean_one(self, hap: Hap):
        def to_clean(hap_arg: Hap) -> bool:
            return hap_arg.hid == hap.hid

        haps_count = self._clean_haps(filter_haps=to_clean)
        logger.debug(f"Deleted {haps_count} haps")

    def clean(self, clean_all: bool = False):
        def to_clean(hap: Hap) -> bool:
            return hap.status == Status.SUCCESS or (
                hap.status == Status.FAILED and clean_all
            )

        haps_count = self._clean_haps(filter_haps=to_clean)

        if haps_count:
            self.ui.print(
                f"{config.ICON_INFO} Deleted {haps_count} finished haps",
                style=f"{config.COLOR_MAIN} bold",
            )
        else:
            self.ui.error("Nothing to clean")

    def kill(self, haps: List[Hap], verbose: bool = True) -> int:
        killed_counter = 0
        for hap in haps:
            if hap.active:
                logger.info(f"Killing {hap}...")
                kill_proc_tree(hap.pid)
                killed_counter += 1

        if killed_counter and verbose:
            self.ui.print(
                f"{config.ICON_KILLED} Killed {killed_counter} active haps",
                style=f"{config.COLOR_MAIN} bold",
            )
        elif verbose:
            self.ui.error("No active haps to kill")
        return killed_counter

    def signal(self, hap: Hap, sig: signal.Signals):
        if hap.active:
            sig_text = (
                f"[bold]{sig.name}[/] ([{config.COLOR_MAIN}]{signal.strsignal(sig)}[/])"
            )
            self.ui.print(f"{config.ICON_INFO} Sending {sig_text} to hap {hap}")
            hap.proc.send_signal(sig)
        else:
            self.ui.error("Cannot send signal to the inactive hap")

    def restart(self, hap: Hap) -> None:
        hid, name, cmd, workdir, restarts, redirect_stderr = (
            hap.hid,
            hap.name,
            hap.cmd,
            hap.workdir,
            hap.restarts,
            hap.redirect_stderr,
        )
        if hap.active:
            self.kill([hap], verbose=False)

        hap_killed = self.get_hap(hid)
        while hap_killed.active:
            # NOTE: re-read is required as `proc` is a cached property
            hap_killed = self.get_hap(hid)

        rc_exists = wait_created(hap_killed._rc_file, timeout=1)
        if not rc_exists:
            logger.error(
                f"Hap {hap_killed} process was killed, "
                f"but parent did not write return code"
            )

        self._clean_one(hap_killed)

        name = f"{name}{config.RESTART_DELIM}{restarts + 1}"
        self.run_command(
            cmd=cmd,
            workdir=workdir,
            hid=hid,
            name=name,
            redirect_stderr=redirect_stderr,
        )

    def rename_hap(self, hap: Hap, new_name: str):
        rich_text = (
            f"{config.ICON_INFO} Renamed [{config.COLOR_ACCENT}]{hap.name}[/] "
            f"to [{config.COLOR_MAIN} bold]{new_name}[/]"
        )
        if hap.restarts:
            new_name = f"{new_name}{config.RESTART_DELIM}{hap.restarts}"
        hap.set_name(new_name)
        self.ui.print(rich_text)



import sys

import click

from hapless import Status
from hapless.cli_utils import get_or_exit, hap_argument, hapless
from hapless.utils import logger


@click.group(invoke_without_command=True)
@hap_argument
def wrap_hap(hap_alias: str) -> None:
    hap = get_or_exit(hap_alias)
    if hap.status != Status.UNBOUND:
        message = f"Hap {hap} has to be unbound, found instead {hap.status}\n"
        with open(hap.stderr_path, "a") as f:
            f.write(message)
        logger.error(message)
        sys.exit(1)
    hapless._wrap_subprocess(hap)


if __name__ == "__main__":
    wrap_hap()



import json
import os
import pwd
import random
import string
import time
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, Optional, Union

try:
    from functools import cached_property
except ImportError:
    # Fallback for Python 3.7
    from backports.cached_property import cached_property

import humanize
import psutil

from hapless import config
from hapless.utils import allow_missing, get_mtime, logger


class Status(str, Enum):
    # Created status
    UNBOUND = "unbound"
    # Active statuses
    PAUSED = "paused"
    RUNNING = "running"
    # Finished statuses
    FAILED = "failed"
    SUCCESS = "success"


class Hap(object):
    def __init__(
        self,
        hap_path: Path,
        *,
        name: Optional[str] = None,
        cmd: Optional[str] = None,
        workdir: Optional[Union[str, Path]] = None,
        redirect_stderr: bool = False,
    ) -> None:
        if not hap_path.is_dir():
            raise ValueError(f"Path {hap_path} is not a directory")

        self._hap_path = hap_path
        self._hid: str = hap_path.name

        self._pid_file = hap_path / "pid"
        self._rc_file = hap_path / "rc"
        self._name_file = hap_path / "name"
        self._cmd_file = hap_path / "cmd"
        self._workdir_file = hap_path / "workdir"
        self._env_file = hap_path / "env"

        self._stdout_path = hap_path / "stdout.log"
        self._stderr_path = hap_path / "stderr.log"

        self._set_logfiles(redirect_stderr)
        self._set_raw_name(name)
        self._set_command_context(cmd, workdir)

    def set_name(self, name: str):
        with open(self._name_file, "w") as f:
            f.write(name)

    def set_return_code(self, rc: int):
        with open(self._rc_file, "w") as f:
            f.write(f"{rc}")

    def _set_raw_name(self, raw_name: Optional[str]):
        """
        Set name for the first time on hap creation.
        """
        if raw_name is None:
            suffix = self.get_random_name()
            raw_name = f"hap-{suffix}"

        if self.raw_name is None:
            self.set_name(raw_name)

    def _set_command_context(
        self,
        cmd: Optional[str],
        workdir: Optional[Union[str, Path]],
    ) -> None:
        """
        Set command and working directory for the first time on hap creation.
        """
        if self.cmd is None:
            if cmd is None:
                raise ValueError("Command to run is not provided")
            with open(self._cmd_file, "w") as f:
                f.write(cmd)

        if self.workdir is None:
            workdir = Path(workdir or os.getcwd())
            # NOTE: should be ValueError, but we need defaults to keep
            # compatibility with the older versions
            if not workdir.exists() or not workdir.is_dir():
                raise ValueError("Workdir should be a path to existing directory")

            with open(self._workdir_file, "w") as f:
                f.write(f"{workdir}")

    def _set_pid(self, pid: int):
        with open(self._pid_file, "w") as pid_file:
            pid_file.write(f"{pid}")

        if not psutil.pid_exists(pid):
            raise RuntimeError(f"Process with pid {pid} is gone")

    def _set_logfiles(self, redirect_stderr: bool):
        if redirect_stderr:
            logger.debug("Process stderr will be redirected to stdout file")
        if not self._stdout_path.exists() and not redirect_stderr:
            self._stderr_path.touch(exist_ok=True, mode=0o644)
        self._stdout_path.touch(exist_ok=True, mode=0o644)

    def _set_env(self):
        proc = self.proc

        environ = {}
        if proc is not None:
            try:
                environ = proc.environ()
            except (ProcessLookupError, psutil.NoSuchProcess) as e:
                logger.error(f"Cannot get environment: {e}")

        if not environ:
            logger.warning(
                "Cannot get environment from the process. "
                "Fallback to current environment"
            )
            environ = dict(os.environ)

        with open(self._env_file, "w") as env_file:
            env_file.write(json.dumps(environ))

    def bind(self, pid: int):
        """
        Associate hap object with existing process by pid.
        """
        try:
            self._set_pid(pid)
            self._set_env()
        except (RuntimeError, psutil.AccessDenied) as e:
            logger.error(f"Cannot bind due to {e}")

    @staticmethod
    def get_random_name(length: int = 6) -> str:
        return "".join(
            random.sample(
                string.ascii_lowercase + string.digits,
                length,
            )
        )

    # TODO: add extended status to show panel proc.status()
    @property
    def status(self) -> Status:
        if self.pid is None and self.rc is None:
            # No existing process or no return code from the finished one
            return Status.UNBOUND

        proc = self.proc
        if proc is not None:
            if proc.status() == psutil.STATUS_STOPPED:
                return Status.PAUSED
            return Status.RUNNING

        if self.rc != 0:
            return Status.FAILED

        return Status.SUCCESS

    @cached_property
    def proc(self):
        # NOTE: this is cached for the instance lifetime, fits our use case
        if self.pid is None:
            return

        try:
            return psutil.Process(self.pid)
        except psutil.NoSuchProcess as e:
            logger.warning(f"Cannot find process: {e}")

    @property
    @allow_missing
    def cmd(self) -> Optional[str]:
        with open(self._cmd_file) as f:
            return f.read()

    @property
    @allow_missing
    def workdir(self) -> Optional[Path]:
        with open(self._workdir_file) as f:
            return Path(f.read())

    @property
    @allow_missing
    def rc(self) -> Optional[int]:
        with open(self._rc_file) as f:
            return int(f.read())

    @property
    def runtime(self) -> str:
        proc = self.proc
        runtime = 0
        if proc is not None:
            runtime = time.time() - proc.create_time()
        elif self._pid_file.exists():
            start_time = get_mtime(self._pid_file)
            finish_time = get_mtime(self._rc_file) or get_mtime(self.stderr_path)
            runtime = finish_time - start_time

        return humanize.naturaldelta(runtime)

    @property
    def start_time(self) -> Optional[str]:
        if self._pid_file.exists():
            return datetime.fromtimestamp(os.path.getmtime(self._pid_file)).strftime(
                config.DATETIME_FORMAT
            )
        logger.info("No start time as hap has not started yet")

    @property
    def end_time(self) -> Optional[str]:
        if self._rc_file.exists():
            return datetime.fromtimestamp(os.path.getmtime(self._rc_file)).strftime(
                config.DATETIME_FORMAT
            )
        logger.info("No end time as hap has not finished yet")

    @property
    def active(self) -> bool:
        return self.proc is not None

    @property
    def hid(self) -> str:
        return self._hid

    @property
    @allow_missing
    def pid(self) -> Optional[int]:
        with open(self._pid_file) as f:
            return int(f.read())

    @property
    @allow_missing
    def env(self) -> Dict[str, str]:
        proc = self.proc
        if proc is not None:
            return proc.environ()

        with open(self._env_file) as f:
            return json.loads(f.read())

    @property
    @allow_missing
    def raw_name(self) -> Optional[str]:
        with open(self._name_file) as f:
            return f.read().strip()

    @cached_property
    def name(self) -> str:
        """
        Base name without restarts counter.
        """
        return self.raw_name.split(config.RESTART_DELIM)[0]

    @cached_property
    def restarts(self) -> int:
        _, *rest = self.raw_name.rsplit("@", maxsplit=1)
        return int(rest[0]) if rest else 0

    @property
    def path(self) -> Path:
        return self._hap_path

    @property
    def stdout_path(self) -> Path:
        return self._stdout_path

    @property
    def stderr_path(self) -> Path:
        if self.redirect_stderr:
            return self._stdout_path
        return self._stderr_path

    @property
    def redirect_stderr(self) -> bool:
        return not self._stderr_path.exists()

    @property
    def accessible(self) -> bool:
        """
        Check if current user has control over the hap.
        NOTE: EAFP is preferable here instead
        https://docs.python.org/3/library/os.html#os.access
        """
        return os.access(self.path, os.F_OK | os.R_OK | os.W_OK | os.X_OK)

    @property
    def owner(self) -> str:
        stat = self.path.stat()
        try:
            owner = pwd.getpwuid(stat.st_uid).pw_name
        except KeyError:
            owner = f"{stat.st_uid}:{stat.st_gid}"
        return owner

    def serialize(self) -> dict:
        """
        Serialize hap object into a dictionary.
        """
        return {
            "hid": self.hid,
            "name": self.name,
            "pid": str(self.pid) if self.pid is not None else None,
            "rc": str(self.rc) if self.rc is not None else None,
            "cmd": self.cmd,
            "workdir": str(self.workdir),
            "status": self.status.value,
            "runtime": self.runtime,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "restarts": str(self.restarts),
            "stdout_file": str(self.stdout_path),
            "stderr_file": str(self.stderr_path),
        }

    def __str__(self) -> str:
        return f"#{self.hid} ({self.name})"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self} object at {hex(id(self))}>"

    def __rich__(self) -> str:
        pid_text = f"with PID [[{config.COLOR_MAIN} bold]{self.pid}[/]]"
        rich_text = (
            f"hap {config.ICON_HAP}{self.hid} "
            f"([{config.COLOR_MAIN} bold]{self.name}[/])"
        )
        if self.pid:
            return f"{rich_text} {pid_text}"
        return rich_text



import logging
import os
import shlex
import signal
import time
from contextlib import nullcontext
from functools import wraps
from pathlib import Path
from typing import Callable, Optional, TypeVar

import click
import psutil
import structlog
from rich.spinner import Spinner
from rich.text import Text
from typing_extensions import ParamSpec

from hapless import config

P = ParamSpec("P")
R = TypeVar("R")


def allow_missing(func: Callable[P, R]) -> Callable[P, Optional[R]]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[R]:
        try:
            return func(*args, **kwargs)
        except FileNotFoundError:
            pass

    return wrapper


def shlex_join_backport(split_command):
    """Return a shell-escaped string"""
    return " ".join(shlex.quote(arg) for arg in split_command)


class timed(object):
    def __init__(self):
        self._start = time.time()
        self._end = None

    @property
    def elapsed(self):
        if self._end is not None:
            return self._end - self._start

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._end = time.time()
        return self


class dummy_live(nullcontext):
    def update(self, *args, **kwargs):
        pass


def wait_created(
    path: Path,
    interval: float = 0.1,
    timeout: float = config.FAILFAST_TIMEOUT,
    *,
    live_context=dummy_live(),
) -> bool:
    start = time.time()
    with live_context:
        elapsed = 0
        while not path.exists() and elapsed < timeout:
            elapsed = time.time() - start
            spinner = Spinner(
                "dots",
                text=Text.from_markup(
                    f"checking process health for "
                    f"[bold {config.COLOR_MAIN}]{int(timeout - elapsed)}s[/]..."
                ),
                style=f"{config.COLOR_MAIN}",
            )
            live_context.update(spinner)
            time.sleep(interval)
    return path.exists()


def validate_signal(ctx, param, value):
    try:
        signal_code = int(value)
    except ValueError:
        raise click.BadParameter("Signal should be a valid integer value")

    try:
        return signal.Signals(signal_code)
    except ValueError:
        raise click.BadParameter(f"{signal_code} is not a valid signal code")


def kill_proc_tree(pid, sig=signal.SIGKILL, include_parent=True):
    if pid == os.getpid():
        raise ValueError("Would not kill myself")

    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        logger.warning(f"Process {pid} is already gone, nothing to kill")
        return

    children = parent.children(recursive=True)
    if include_parent:
        children.append(parent)
    for p in children:
        try:
            p.send_signal(sig)
            logger.debug(f"Sent {sig} to {p.pid} process")
        except psutil.NoSuchProcess:
            pass


def get_mtime(path: Path) -> Optional[float]:
    if path.exists():
        return os.path.getmtime(path)


def configure_logger() -> logging.Logger:
    logger = structlog.get_logger()
    level = logging.DEBUG if config.DEBUG else logging.CRITICAL
    structlog.configure(wrapper_class=structlog.make_filtering_bound_logger(level))
    return logger


logger = configure_logger()



from typing import List, Optional

from rich.console import Console
from rich.live import Live

from hapless import config
from hapless.formatters import Formatter, TableFormatter
from hapless.hap import Hap


class ConsoleUI:
    def __init__(self, disable: bool = False) -> None:
        self.console = Console(highlight=False, quiet=disable)
        self.default_formatter = TableFormatter()
        self.disable = disable

    def print(self, *args, **kwargs):
        return self.console.print(*args, **kwargs)

    def print_plain(self, text: str):
        return self.console.print(
            text,
            markup=False,
            highlight=False,
            emoji=False,
            no_wrap=True,
            overflow="ignore",
            crop=False,
        )

    def error(self, message: str):
        return self.console.print(
            f"{config.ICON_INFO} {message}",
            style=f"{config.COLOR_ERROR} bold",
            overflow="ignore",
            crop=False,
        )

    def get_live(self):
        return Live(
            console=self.console,
            refresh_per_second=10,
            transient=True,
        )

    def stats(self, haps: List[Hap], formatter: Optional[Formatter] = None):
        if not haps:
            self.console.print(
                f"{config.ICON_INFO} No haps are currently running",
                style=f"{config.COLOR_MAIN} bold",
            )
            return
        formatter = formatter or self.default_formatter
        haps_data = formatter.format_list(haps)
        self.console.print(haps_data, soft_wrap=True)

    def show_one(self, hap: Hap, formatter: Optional[Formatter] = None):
        formatter = formatter or self.default_formatter
        hap_data = formatter.format_one(hap)
        self.console.print(hap_data, soft_wrap=True)



from pathlib import Path
from typing import Optional

import environ

env = environ.Env()

DEBUG = env.bool("HAPLESS_DEBUG", default=False)

HAPLESS_DIR: Optional[Path] = env("HAPLESS_DIR", cast=Path, default=None)

COLOR_MAIN = "#fdca40"
COLOR_ACCENT = "#3aaed8"
COLOR_ERROR = "#f64740"
STATUS_COLORS = {
    "running": "#f79824",
    "paused": "#f6efee",
    "success": "#4aad52",
    "failed": COLOR_ERROR,
}

ICON_HAP = "⚡️"
ICON_INFO = "🧲"
ICON_STATUS = "•"
ICON_KILLED = "💀"

FAILFAST_TIMEOUT = env.int("HAPLESS_FAILFAST_TIMEOUT", default=5)
DATETIME_FORMAT = "%H:%M:%S %Y/%m/%d"
TRUNCATE_LENGTH = 36
RESTART_DELIM = "@"

NO_FORK = env.bool("HAPLESS_NO_FORK", default=False)

REDIRECT_STDERR = env.bool("HAPLESS_REDIRECT_STDERR", default=False)
